from collections import defaultdict

from neurobench.metrics.abstract.workload_metric import (
    WorkloadMetric,
    AccumulatedMetric,
)
from neurobench.metrics import workload as workload_metrics
from neurobench.metrics.utils.utils import convert_to_class_name

from neurobench.models import NeuroBenchModel
from torch import Tensor
from typing import List, Dict


class WorkloadMetricManager:
    """Orchestrator for managing and executing workload metrics on a model."""

    def __init__(self, metric_list: List):
        """
        Initializes the orchestrator with a list of workload metrics.

        Args:
            metric_list (list): A list of metric identifiers, either as strings corresponding to internal
                                metric names or as custom metric objects.

        Raises:
            ValueError: If a string-based metric is not found in the internal workload_metrics module.
            TypeError: If a custom metric object does not inherit from WorkloadMetric.

        """
        self.metrics = {}

        # Store workload metrics
        for item in metric_list:
            if isinstance(item, str):
                class_name = convert_to_class_name(item)
                metric_class = getattr(workload_metrics, class_name, None)
                print(
                    f"[DEPRECATION WARNING]: Using string-based metric names ('{item}') is deprecated "
                    "and will be removed in a future release. "
                    "Please update your code to use the corresponding metric class directly instead. "
                    f"For example, replace the string name with the class '{class_name}'."
                )
                if metric_class is None:
                    raise ValueError(
                        f"Metric '{item}' not found in the 'workload_metrics' module."
                    )
                if not issubclass(metric_class, WorkloadMetric):
                    raise TypeError(
                        f"The metric class '{item}' must inherit from WorkloadMetric."
                    )
                self.metrics[item] = metric_class()
            else:
                # Validate custom metric object
                if not issubclass(item, WorkloadMetric):
                    raise TypeError(
                        f"Custom metric '{item.__name__}' must inherit from WorkloadMetric."
                    )
                self.metrics[item.__name__]: WorkloadMetric = item()

        # Track whether hooks are required
        self.requires_hooks = any(
            metric.requires_hooks for metric in self.metrics.values()
        )

        self.results = defaultdict(float)

    def clean_results(self) -> None:
        """Reset the results dictionary to zero."""
        self.results = defaultdict(float)

    def register_hooks(self, model: NeuroBenchModel) -> None:
        """
        Register hooks on the model for metrics that require them.

        Args:
            model: The model on which the hooks will be registered.

        """
        if self.requires_hooks:
            model.register_hooks()

    def reset_hooks(self, model: NeuroBenchModel) -> None:
        """
        Cleanup hooks on the model after metrics have been run.

        Args:
            model: The model on which the hooks will be cleaned up.

        """
        if self.requires_hooks:
            model.reset_hooks()

    def cleanup_hooks(self, model: NeuroBenchModel) -> None:
        """
        Cleanup hooks on the model after metrics have been run.

        Args:
            model: The model on which the hooks will be cleaned up.

        """
        if self.requires_hooks:
            model.cleanup_hooks()

    def close_hooks(self, model: NeuroBenchModel) -> None:
        """Close hooks on the model after metrics have been run."""
        if self.requires_hooks:
            model.close_hooks()

    def initialize_metrics(self) -> None:
        """Initialize the metrics for a new workload run."""

        for m in self.metrics.keys():
            if isinstance(self.metrics[m], AccumulatedMetric):
                self.metrics[m].reset()

    def run_metrics(
        self,
        model: NeuroBenchModel,
        preds: Tensor,
        data: tuple[Tensor, Tensor],
        batch_size: int,
        dataset_len: int,
    ) -> Dict:
        """
        Executes all workload metrics on the provided model and data.

        Args:
            model: The model on which the metrics will be run.
            preds: A tensor of model predictions.
            data: A tuple of inputs and targets.
            batch_size: The size of the batch.
            dataset_len: The length of the dataset.

        Returns:
            Dict: A dictionary where the keys are metric names and the values are the results of the metric calculations.

        """

        for name, metric in self.metrics.items():
            try:
                result = metric(model, preds, data)

                # Handle AccumulatedMetric separately
                if isinstance(metric, AccumulatedMetric):
                    self.results[name] = result
                else:
                    # Accumulate results, normalizing by dataset length
                    self.results[name] += (result * batch_size) / dataset_len

            except Exception as e:
                print(f"Error running workload metric '{name}': {e}")
                self.results[name] = 0.0  # Graceful fallback in case of errors

        return self.results



from abc import ABC, abstractmethod
from torch import Tensor


class NeuroBenchPostProcessor(ABC):
    """
    Abstract class for NeuroBench postprocessors.

    Postprocessors take the spiking output from the models and provide several methods
    of combining them. Individual postprocessors are responsible for implementing init
    and call functions.

    """

    @abstractmethod
    def __call__(self, spikes: Tensor) -> Tensor:
        """
        Process tensor of spiking data of format (batch, timesteps, ...) to match spikes
        to ground truth.

        Args:
            spikes (Tensor): A torch tensor of spikes output by a NeuroBenchModel of
                shape (batch, timestep, ...)
        Returns:
            Tensor: A tensor of shape (batch, ...) with the processed spikes

        """



from abc import ABC, abstractmethod
from neurobench.hooks.neuron import NeuronHook
from neurobench.hooks.layer import LayerHook
from neurobench.blocks.activation import SUPPORTED_ACTIVATIONS
from neurobench.blocks.layer import SUPPORTED_LAYERS


class NeuroBenchModel(ABC):
    """
    Abstract class for NeuroBench models.

    Individual model frameworks are responsible for defining model inference.

    """

    def __init__(self):
        """
        Init using a trained network.

        Args:
            net: A trained network

        """
        self.activation_modules = list(SUPPORTED_ACTIVATIONS)
        self.activation_hooks = []
        self.connection_hooks = []
        self.first_layer = None

        # self.supported_layers = (
        #     nn.Linear,
        #     nn.Conv2d,
        #     nn.Conv1d,
        #     nn.Conv3d,
        #     nn.RNNBase,
        #     nn.RNNCellBase,
        # )

    @abstractmethod
    def __call__(self, batch):
        """
        Includes the whole pipeline from data to inference (output should be same format
        as targets).

        Args:
            batch: A batch of data to run inference on

        """
        pass

    @abstractmethod
    def __net__(self):
        """Returns the underlying network."""
        pass

    def set_first_layer(self, layer):
        """Sets the first layer of the network."""
        self.first_layer = layer

    def add_activation_module(self, activaton_module):
        """Add a cutomized activaton_module that can be detected after running the
        preprocessing pipeline for detecting activation functions."""
        self.activation_modules.append(activaton_module)

    def activation_layers(self):
        """
        Retrieve all activation layers in the network, including spiking neurons.

        Returns:
            list: Activation layers.

        """

        def is_activation_layer(module):
            """Check if a module is an activation layer."""
            return any(
                isinstance(module, act_mod) for act_mod in self.activation_modules
            )

        def find_activation_layers(module):
            """Recursively find activation layers in a module."""
            layers = []
            for child_name, child in module.named_children():
                if is_activation_layer(child):
                    layers.append({"layer_name": child_name, "layer": child})
                elif list(child.children()):  # Check for nested submodules
                    layers.extend(find_activation_layers(child))
            return layers

        return find_activation_layers(self.__net__())

    def connection_layers(self):
        """
        Retrieve all connection layers in the network.

        Connection layers include Linear, Conv, and RNN-based layers.

        Returns:
            list: Connection layers.

        """

        def find_connection_layers(module):
            """Recursively find connection layers in a module."""
            layers = []
            for child in module.children():
                if isinstance(child, SUPPORTED_LAYERS):
                    layers.append(child)
                elif list(child.children()):  # Check for nested submodules
                    layers.extend(find_connection_layers(child))
            return layers

        return find_connection_layers(self.__net__())

    def reset_hooks(self):
        """Resets all the hooks (activation hooks and connection hooks) in parallel."""
        for hook in self.activation_hooks + self.connection_hooks:
            hook.reset()

    def close_hooks(self):
        """Closes all the hooks (activation hooks and connection hooks)"""
        for hook in self.activation_hooks + self.connection_hooks:
            hook.close()

    def cleanup_hooks(self):
        """Closes all the hooks (activation hooks and connection hooks)"""
        for hook in self.activation_hooks + self.connection_hooks:
            hook.reset()
            hook.close()

        self.activation_hooks.clear()
        self.connection_hooks.clear()

    def register_hooks(self):
        """Registers hooks for the model."""

        # Registered activation hooks
        for layer in self.activation_layers():
            layer_name = layer["layer_name"]
            layer = layer["layer"]
            self.activation_hooks.append(NeuronHook(layer, layer_name))

        for layer in self.connection_layers():
            self.connection_hooks.append(LayerHook(layer))



import torch
from torch import Tensor
from neurobench.metrics.utils.decorators import check_shapes

from neurobench.metrics.abstract.workload_metric import WorkloadMetric


class ClassificationAccuracy(WorkloadMetric):
    """Classification accuracy of the model predictions."""

    def __init__(self):
        """Initialize the ClassificationAccuracy metric."""
        super().__init__(requires_hooks=False)

    @check_shapes
    def __call__(self, model, preds, data):
        """
        Compute classification accuracy.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: Classification accuracy

        """

        equal = torch.eq(preds, data[1])
        return torch.mean(equal.float()).item()



from .membrane_updates import MembraneUpdates
from .activation_sparsity import ActivationSparsity
from .synaptic_operations import SynapticOperations
from .classification_accuracy import ClassificationAccuracy
from .activation_sparsity_by_layer import ActivationSparsityByLayer
from .mse import MSE
from .smape import SMAPE
from .r2 import R2
from .coco_map import CocoMap

__stateless__ = [
    "ClassificationAccuracy",
    "ActivationSparsity",
    "MSE",
    "SMAPE",
    "ActivationSparsityByLayer",
]

__stateful__ = ["MembraneUpdates", "SynapticOperations", "R2", "CocoMap"]

__all__ = __stateful__ + __stateless__



from neurobench.processors.abstract.postprocessor import NeuroBenchPostProcessor


class ChooseMaxCount(NeuroBenchPostProcessor):
    """Returns the class with the highest spike count over the sample."""

    def __call__(self, spikes):
        """
        Returns the class with the highest spike count over the sample.

        Args:
            spikes (Tensor): A torch tensor of spikes of shape (batch, timestep, classes)

        Returns:
            Tensor: A torch tensor of shape (batch,) with the class index of the highest spike count

        """
        return spikes.sum(1).argmax(1)



from abc import ABC, abstractmethod
from torch import Tensor


class NeuroBenchPreProcessor(ABC):
    """
    Abstract class for NeuroBench pre-processors.

    Individual pre-processors are responsible for implementing init and call functions.

    """

    @abstractmethod
    def __init__(self, *args):
        """
        Initialize pre-processor with any parameters needed.

        Args:
            args: Any arguments needed for pre-processing.

        """

    @abstractmethod
    def __call__(self, dataset: tuple[Tensor, Tensor]) -> tuple[Tensor, Tensor]:
        """
        Process dataset of format (data, targets), or (data, targets, kwargs) to prepare
        for model inference.

        Args:
            dataset (tuple): A tuple of (data, targets) or (data, targets, kwargs) where data is a PyTorch tensor of shape (batch, timesteps, ...)

        Returns:
            tuple: A tuple of (data, targets) or (data, targets, kwargs) where data is a PyTorch tensor of shape (batch, timesteps, ...)

        """



from abc import ABC
from neurobench.processors.abstract import (
    NeuroBenchPreProcessor,
    NeuroBenchPostProcessor,
)
from typing import List, Callable, Tuple
from torch import Tensor


class ProcessorManager(ABC):
    """
    Manager class for handling preprocessors and postprocessors.

    This class manages the list of preprocessing and postprocessing steps, ensuring
    that the appropriate processors are applied to data before and after the model
    processing. It supports replacing the lists of preprocessors and postprocessors
    and ensures that each processor is of the correct type.

    Attributes:
        preprocessors (List[NeuroBenchPreProcessor]): List of preprocessors to
        be applied before model inference.
        postprocessors (List[NeuroBenchPostProcessor]): List of postprocessors
        to be applied after model inference.

    """

    def __init__(
        self,
        preprocessors: List[
            NeuroBenchPreProcessor
            | Callable[[Tuple[Tensor, Tensor]], Tuple[Tensor, Tensor]]
        ],
        postprocessors: List[NeuroBenchPostProcessor | Callable[[Tensor], Tensor]],
    ):
        """
        Initialize the ProcessorManager with the given preprocessors and postprocessors.

        Args:
            preprocessors (List[NeuroBenchPreProcessor]): List of preprocessors
            to be applied to the data.
            postprocessors (List[NeuroBenchPostProcessor]): List of postprocessors
            to be applied to the data.

        Raises:
            TypeError: If any element in preprocessors is not an instance of
            NeuroBenchPreProcessor, or if any element in postprocessors is not
            an instance of NeuroBenchPostProcessor.

        """

        if any(
            not (isinstance(p, NeuroBenchPreProcessor) or callable(p))
            for p in preprocessors
        ):
            raise TypeError(
                "All preprocessors must be instances of NeuroBenchPreProcessor or callable functions"
            )
        if any(
            not (isinstance(p, NeuroBenchPostProcessor) or callable(p))
            for p in postprocessors
        ):
            raise TypeError(
                "All postprocessors must be instances of NeuroBenchPostProcessor or callable functions"
            )

        self.preprocessors = preprocessors
        self.postprocessors = postprocessors

    def replace_preprocessors(
        self,
        preprocessors: List[
            NeuroBenchPreProcessor
            | Callable[[Tuple[Tensor, Tensor]], Tuple[Tensor, Tensor]]
        ],
    ):
        """
        Replace the current list of preprocessors with the provided list.

        Args:
            preprocessors (List[NeuroBenchPreProcessor]): List of new preprocessors
            to replace the current ones.

        Raises:
            TypeError: If any element in the provided list is not an instance of
            NeuroBenchPreProcessor.

        """
        if any(
            not (isinstance(p, NeuroBenchPreProcessor) or callable(p))
            for p in preprocessors
        ):
            raise TypeError(
                "All preprocessors must be instances of NeuroBenchPreProcessor or callable functions"
            )
        self.preprocessors = preprocessors

    def replace_postprocessors(
        self, postprocessors: List[NeuroBenchPostProcessor | Callable[[Tensor], Tensor]]
    ):
        """
        Replace the current list of postprocessors with the provided list.

        Args:
            postprocessors (List[NeuroBenchPostProcessor]): List of new postprocessors
            to replace the current ones.

        Raises:
            TypeError: If any element in the provided list is not an instance of
            NeuroBenchPostProcessor.

        """
        if any(
            not (isinstance(p, NeuroBenchPostProcessor) or callable(p))
            for p in postprocessors
        ):
            raise TypeError(
                "All postprocessors must be instances of NeuroBenchPostProcessor or callable functions"
            )
        self.postprocessors = postprocessors

    def preprocess(self, data: Tuple[Tensor, Tensor]) -> tuple[Tensor, Tensor]:
        """
        Apply preprocessing steps to the input data.

        Applies each preprocessor in the list sequentially to the provided data.

        Args:
            data (tuple[Tensor, Tensor]): The data to preprocess.

        Returns:
            tuple[Tensor, Tensor]: The preprocessed data.

        """
        for preprocessor in self.preprocessors:
            data = preprocessor(data)
        return data

    def postprocess(self, data: Tensor) -> Tensor:
        """
        Apply postprocessing steps to the input data.

        Applies each postprocessor in the list sequentially to the provided data.

        Args:
            data (Tensor): The data to postprocess.

        Returns:
            Tensor: The postprocessed data.

        """
        for postprocessor in self.postprocessors:
            data = postprocessor(data)
        return data



from abc import ABC


class LayerHook(ABC):
    def __init__(self, layer) -> None:
        self.layer = layer
        self.inputs = []
        if layer is not None:
            self.hook = layer.register_forward_pre_hook(self.hook_fn)
        else:
            self.hook = None

    def hook_fn(self, module, input):
        self.inputs.append(input)

    def register_hook(self):
        self.hook = self.layer.register_forward_pre_hook(self.hook_fn)

    def reset(self):
        self.inputs.clear()

    def close(self):
        if self.hook:
            self.hook.remove()



import torch
from torch import Tensor
from neurobench.metrics.utils.decorators import check_shapes

from neurobench.metrics.abstract.workload_metric import WorkloadMetric


class MSE(WorkloadMetric):
    """Mean squared error of the model predictions."""

    def __init__(self):
        """Initialize the MSE metric."""

        super().__init__(requires_hooks=False)

    @check_shapes
    def __call__(self, model, preds: Tensor, data: Tensor) -> float:
        """
        Compute mean squared error.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: Mean squared error.

        """

        return torch.mean((preds - data[1]) ** 2).item()



from neurobench.metrics.abstract import StaticMetric
from neurobench.blocks.layer import STATELESS_LAYERS, RECURRENT_LAYERS, RECURRENT_CELLS
import torch
from snntorch import SpikingNeuron


class ConnectionSparsity(StaticMetric):
    """
    Sparsity of model connections between layers.

    Based on number of zeros
    in supported layers, other layers are not taken into account in the computation:
    Supported layers:
    Linear
    Conv1d, Conv2d, Conv3d
    RNN, RNNBase, RNNCell
    LSTM, LSTMBase, LSTMCell
    GRU, GRUBase, GRUCell

    """

    def __call__(self, model):
        """
        Compute connection sparsity.

        Args:
            model: A NeuroBenchModel.
        Returns:
            float: Connection sparsity, rounded to 3 decimals.

        """

        def get_nr_zeros_weights(module):
            """
            Get the number of zeros in a module's weights.

            Args:
                module: A torch.nn.Module.
            Returns:
                int: Number of zeros in the module's weights.

            """
            children = list(module.children())
            if len(children) == 0:  # it is a leaf
                # print(module)
                if isinstance(module, STATELESS_LAYERS):
                    count_zeros = torch.sum(module.weight == 0)
                    count_weights = module.weight.numel()
                    return count_zeros, count_weights

                elif isinstance(module, RECURRENT_LAYERS):
                    attribute_names = []
                    for i in range(module.num_layers):
                        param_names = ["weight_ih_l{}{}", "weight_hh_l{}{}"]
                        if module.bias:
                            param_names += ["bias_ih_l{}{}", "bias_hh_l{}{}"]
                        if module.proj_size > 0:  # it is lstm
                            param_names += ["weight_hr_l{}{}"]

                        attribute_names += [x.format(i, "") for x in param_names]
                        if module.bidirectional:
                            suffix = "_reverse"
                            attribute_names += [
                                x.format(i, suffix) for x in param_names
                            ]

                    count_zeros = 0
                    count_weights = 0
                    for attr in attribute_names:
                        attr_val = getattr(module, attr)
                        count_zeros += torch.sum(attr_val == 0)
                        count_weights += attr_val.numel()

                    return count_zeros, count_weights

                elif isinstance(module, RECURRENT_CELLS):
                    attribute_names = ["weight_ih", "weight_hh"]
                    if module.bias:
                        attribute_names += ["bias_ih", "bias_hh"]

                    count_zeros = 0
                    count_weights = 0
                    for attr in attribute_names:
                        attr_val = getattr(module, attr)
                        count_zeros += torch.sum(attr_val == 0)
                        count_weights += attr_val.numel()

                    return count_zeros, count_weights

                elif isinstance(module, SpikingNeuron):
                    return 0, 0  # it is a neuromorphic neuron layer
                else:
                    # print('Module type: ', module, 'not found.')
                    return 0, 0

            else:
                count_zeros = 0
                count_weights = 0
                for child in children:
                    child_zeros, child_weights = get_nr_zeros_weights(child)
                    count_zeros += child_zeros
                    count_weights += child_weights
                return count_zeros, count_weights

        # Pull the layers from the model's network
        layers = model.__net__().children()
        # For each layer, count where the weights are zero
        count_zeros = 0
        count_weights = 0
        for module in layers:
            zeros, weights = get_nr_zeros_weights(module)
            count_zeros += zeros
            count_weights += weights

        # Return the ratio of zeros to weights, rounded to 4 decimals
        return round((count_zeros / count_weights).item(), 4)



import torch
from neurobench.metrics.abstract.workload_metric import AccumulatedMetric
from neurobench.metrics.utils.decorators import check_shapes


class R2(AccumulatedMetric):
    """
    R2 Score of the model predictions.

    Currently implemented for 2D output only.

    """

    def __init__(self):
        """
        Initalize metric state.

        Must hold memory of all labels seen so far.

        """
        super().__init__(requires_hooks=False)
        self.x_sum_squares = 0.0
        self.y_sum_squares = 0.0

        self.x_labels = None
        self.y_labels = None

    def reset(self):
        """Reset metric state."""
        self.x_sum_squares = 0.0
        self.y_sum_squares = 0.0

        self.x_labels = torch.tensor([])
        self.y_labels = torch.tensor([])

    @check_shapes
    def __call__(self, model, preds, data):
        """
        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: R2 Score.
        """
        self.x_sum_squares += torch.sum((data[1][:, 0] - preds[:, 0]) ** 2).item()
        self.y_sum_squares += torch.sum((data[1][:, 1] - preds[:, 1]) ** 2).item()

        if self.x_labels is None:
            self.x_labels = data[1][:, 0]
            self.y_labels = data[1][:, 1]
        else:
            self.x_labels = torch.cat((self.x_labels, data[1][:, 0]))
            self.y_labels = torch.cat((self.y_labels, data[1][:, 1]))

        return self.compute()

    def compute(self):
        """Compute r2 score using accumulated data."""
        x_denom = self.x_labels.var(correction=0) * len(self.x_labels)
        y_denom = self.y_labels.var(correction=0) * len(self.y_labels)

        x_r2 = 1 - (self.x_sum_squares / x_denom)
        y_r2 = 1 - (self.y_sum_squares / y_denom)

        r2 = (x_r2 + y_r2) / 2

        return r2.item()



from neurobench.blocks.layer import STATELESS_LAYERS, RECURRENT_LAYERS, RECURRENT_CELLS
import copy
import torch


def binarize_tensor(tensor, all_ones=False):
    """
    Binarizes a tensor.

    Non-zero entries become 1. If all_ones is True, all entries become 1.

    """
    tensor = tensor.clone()
    tensor[tensor != 0] = 1
    if all_ones:
        tensor[tensor == 0] = 1
    return tensor


def make_binary_copy(layer, all_ones=False):
    """
    Makes a binary copy of the layer.

    Non-zero entries in the layer's weights and biases are set to 1.
    If all_ones is True, all entries (including zeros) are set to 1.

    Args:
        layer (torch.nn.Module): The layer to be binarized.
        all_ones (bool): If True, all entries (including zeros) are set to 1.

    Returns:
        torch.nn.Module: A binary copy of the input layer.

    """
    layer_copy = copy.deepcopy(layer)

    if isinstance(layer, STATELESS_LAYERS):
        layer_copy.weight.data = binarize_tensor(layer_copy.weight.data, all_ones)
        if layer.bias is not None:
            layer_copy.bias.data = binarize_tensor(layer_copy.bias.data, all_ones)

    elif isinstance(layer, RECURRENT_CELLS):
        attribute_names = ["weight_ih", "weight_hh"]
        if layer.bias:
            attribute_names += ["bias_ih", "bias_hh"]

        for attr in attribute_names:
            with torch.no_grad():
                attr_val = getattr(layer_copy, attr)
                setattr(
                    layer_copy,
                    attr,
                    torch.nn.Parameter(binarize_tensor(attr_val.data, all_ones)),
                )

    return layer_copy



from neurobench.metrics import static as static_metrics
from neurobench.metrics.abstract import StaticMetric
from neurobench.metrics.utils.utils import convert_to_class_name
from neurobench.models import NeuroBenchModel
from typing import List, Dict


class StaticMetricManager:
    """Orchestrator for managing and executing static metrics on a model."""

    def __init__(self, metric_list: List):
        """
        Initializes the orchestrator with a list of static metrics.

        Args:
            metric_list (List): A list of metric identifiers, either as strings corresponding to internal
                                metric names or as custom metric objects.

        Raises:
            TypeError: If a metric object does not inherit from StaticMetric.
            ValueError: If a string-based metric is not found in the internal static_metrics module.

        """
        self.metrics = {}

        for item in metric_list:
            if isinstance(item, str):
                class_name = convert_to_class_name(item)
                metric_class = getattr(static_metrics, class_name, None)
                print(
                    f"[DEPRECATION WARNING]: Using string-based metric names ('{item}') is deprecated "
                    "and will be removed in a future release. "
                    "Please update your code to use the corresponding metric class directly instead. "
                    f"For example, replace the string name with the class '{class_name}'."
                )
                if metric_class is None:
                    raise ValueError(
                        f"Metric class '{class_name}' not found in the 'static_metrics' module."
                    )
                if not issubclass(metric_class, StaticMetric):
                    raise TypeError(
                        f"The metric class '{class_name}' must inherit from StaticMetric."
                    )
                self.metrics[item] = metric_class()  # Instantiate the metric class
            else:
                # Validate custom metric object
                if not issubclass(item, StaticMetric):
                    raise TypeError(
                        f"Custom metric '{item.__name__}' must inherit from StaticMetric."
                    )
                self.metrics[item.__name__] = item()

    def run_metrics(self, model: NeuroBenchModel) -> Dict:
        """
        Executes all static metrics on the provided model.

        Args:
            model: The model on which the metrics will be run.

        Returns:
            Dict: A dictionary where the keys are metric names and the values are the results of the metric calculations.

        Raises:
            Exception: If a metric computation fails, it is caught and logged with an error message.

        """
        results = {}
        for name, metric in self.metrics.items():
            try:
                results[name] = metric(model)
            except Exception as e:
                print(f"Error running metric '{name}': {e}")
                results[name] = None  # Graceful fallback in case of errors
        return results



"""
This file contains code based on torch-mate: https://github.com/V0XNIHILI/torch-mate
"""

from typing import List, Optional, Tuple, Union, Dict

import random
import numpy as np
import torch
from torch.utils.data import IterableDataset

from neurobench.datasets.MSWC_dataset import MSWC, MSWC_query


def get_indices_per_class(
    languages,
    root,
    samples_per_class: int,
    support_query_split: Optional[Tuple[int, int]] = None,
) -> Union[Dict[int, List[int]], Dict[int, Tuple[List[int], List[int]]]]:
    indices_per_lang = {}
    for lang in languages:
        indices_per_lang[lang] = {}
        indices_per_class = indices_per_lang[lang]
        dataset = MSWC(root=root, subset="evaluation", language=lang)

        # Ordering samples index per class using the fact that all samples from one class are indexed together
        for i in range(len(dataset) // samples_per_class):
            label = dataset[i * samples_per_class][1]
            indices_per_class[label] = list(
                range(i * samples_per_class, (i + 1) * samples_per_class)
            )

        if support_query_split is not None:
            n_support, n_query = support_query_split

            for key in indices_per_class.keys():
                indices_per_class[key] = (
                    indices_per_class[key][:n_support],
                    indices_per_class[key][n_support : n_support + n_query],
                )
        del dataset

    return indices_per_lang


class IncrementalFewShot(IterableDataset):
    def __init__(
        self,
        k_shot: int,
        root: str,
        inc_languages: list = [
            "fa",
            "eo",
            "pt",
            "eu",
            "pl",
            "cy",
            "nl",
            "ru",
            "es",
            "it",
        ],
        query_shots: int = -1,
        support_query_split: Optional[Tuple[int, int]] = None,
    ):
        """
        Dataset for few shot learning.

        Args:
            k_shot (int, optional): Number of samples per class in the support set.
            root (str): Path of the folder where to find the dataset language folders.
            inc_languages (List[str], optional): List of languages 2 letters names to use as incremental sessions.
            query_shots (Optional[int]): Number of samples per class in the query set. If not set, query_shots is set to k_shot. Defaults to -1.
            support_query_split (Optional[Tuple[int, int]], optional): Create non-overlapping support and query pools of given number of samples per class. Defaults to None.

        """

        self.support_query_split = support_query_split
        self.samples_per_class = 200  # Number of samples per class to use. Used to simplify samples class indexing as the used dataset is ordered (class_0_sample_0, c0s1, c0s2, c1s0, c1s1, c1s2, ...).
        self.indices_per_lang = get_indices_per_class(
            inc_languages, root, self.samples_per_class, self.support_query_split
        )

        self.n_way = 10  # Number of classes in the support and support sets. Fixed based on the dataset
        self.k_shot = k_shot
        self.query_shots = query_shots if query_shots != -1 else k_shot

        self.languages = inc_languages
        self.root = root

    def __iter__(self):
        """
        Get a batch of samples for a k-shot n-way task.

        Returns:
            tuple[List[tuple[Tensor, Tensor]], Dataset, List[int]]: The support as a list of all shots, the query as a Dataset object and the list of cumulative query classes.

        """

        cumulative_classes = {}
        self.cumulative_query = []

        for lang in random.sample(self.languages, len(self.languages)):
            dataset = MSWC(root=self.root, subset="evaluation", language=lang)

            support_classes = random.sample(
                sorted(self.indices_per_lang[lang].keys()), self.n_way
            )
            cumulative_classes[lang] = support_classes

            # Yields iterative sessions
            out = self._inner_iter(lang, dataset, support_classes, cumulative_classes)
            del dataset
            yield out

    def _inner_iter(self, language, dataset, support_classes, cumulative_classes):
        for class_index in cumulative_classes[language]:
            query_indices = np.random.choice(
                self.indices_per_lang[language][class_index][1],
                self.query_shots,
                replace=False,
            )

            self.cumulative_query += [
                (dataset[j][3], class_index, dataset[j][2]) for j in query_indices
            ]

        query_set = MSWC_query(self.cumulative_query)

        X_train_samples = [[] for _ in range(self.k_shot)]
        y_train_samples = [[] for _ in range(self.k_shot)]

        # For every class
        for class_index in support_classes:
            support_indices = np.random.choice(
                self.indices_per_lang[language][class_index][0],
                self.k_shot,
                replace=False,
            )

            # For every (support) shot/sample
            for j, index in enumerate(support_indices):
                data, real_class, _, _ = dataset[index]
                y_train_samples[j].append(real_class)
                X_train_samples[j].append(data)

        support = []
        for x, y in zip(X_train_samples, y_train_samples):
            shot = (torch.stack(x), torch.tensor(y, dtype=torch.long))
            support.append(shot)

        query_classes = []
        for lang_class in cumulative_classes.values():
            query_classes.extend(lang_class)

        out = (support, query_set, query_classes)

        return out



from .neurobench_model import NeuroBenchModel
from .snntorch_models import SNNTorchModel
from .torch_model import TorchModel
from ..utils import _lazy_import

__all__ = ["NeuroBenchModel", "SNNTorchModel", "TorchModel"]



import torch
from neurobench.metrics.abstract.workload_metric import WorkloadMetric


class ActivationSparsity(WorkloadMetric):
    """
    Sparsity of model activations.

    Calculated as the number of zero activations over the total number of activations,
    over all layers, timesteps, samples in data.

    """

    def __init__(self):
        """Initialize the ActivationSparsity metric."""

        super().__init__(requires_hooks=True)

    def __call__(self, model, preds, data):
        """
        Compute activation sparsity.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: Activation sparsity

        """
        # TODO: for a spiking model, based on number of spikes over all timesteps over all samples from all layers
        #       Standard FF ANN depends on activation function, ReLU can introduce sparsity.
        total_spike_num = 0  # Count of non-zero activations
        total_neuro_num = 0  # Count of all activations

        for hook in model.activation_hooks:
            # Skip layers with no outputs
            if not hook.activation_outputs:
                continue

            for (
                spikes
            ) in hook.activation_outputs:  # do we need a function rather than a member
                spike_num, neuro_num = torch.count_nonzero(spikes).item(), torch.numel(
                    spikes
                )
                total_spike_num += spike_num
                total_neuro_num += neuro_num

        # Compute sparsity
        if total_neuro_num == 0:  # Prevent division by zero
            return 0.0

        sparsity = (total_neuro_num - total_spike_num) / total_neuro_num
        return sparsity



from abc import ABC, abstractmethod
from torch import Tensor
from neurobench.models import NeuroBenchModel


class WorkloadMetric(ABC):
    """
    Abstract base class for workload metrics.

    A workload metric is designed to evaluate some aspect of a model's performance
    or behavior, typically during the inference phase, based on its predictions
    and input data. This class defines the basic interface for all workload metrics
    that require computation over batches of data.

    Attributes:
        requires_hooks (bool): Flag indicating if the metric requires hooks for its computation.

    """

    def __init__(self, requires_hooks: bool = False):
        """
        Initialize the WorkloadMetric.

        Args:
            requires_hooks (bool, default=False): Flag indicating if the metric requires hooks

        """
        self._requires_hooks = requires_hooks

    @abstractmethod
    def __call__(
        self, model: NeuroBenchModel, preds: Tensor, data: tuple[Tensor, Tensor]
    ) -> float:
        """
        Compute the workload metric.

        This method must be implemented by any subclass to define how the metric
        should be computed based on the model, predictions, and data.

        Args:
            model (NeuroBenchModel): The model whose performance is being evaluated.
            preds (Tensor): A tensor of model predictions.
            data (tuple[Tensor, Tensor]): A tuple containing the input data (Tensor)
            and the true labels (Tensor).

        Returns:
            float: The computed value of the workload metric.

        """
        pass

    @property
    def requires_hooks(self) -> bool:
        """
        Property indicating whether the metric requires hooks.

        Returns:
            bool: True if the metric requires hooks, False otherwise.

        """
        return self._requires_hooks


class AccumulatedMetric(WorkloadMetric):
    """
    Abstract base class for accumulated workload metrics.

    An accumulated metric tracks values over multiple batches or iterations and computes
    the final metric value after accumulating data. It extends the WorkloadMetric class
    and adds functionality for resetting and computing the accumulated metric over time.

    """

    def __init__(self, requires_hooks: bool = False):
        """
        Initialize the AccumulatedMetric.

        Args:
            requires_hooks (bool, default=False): Flag indicating if the metric requires hooks

        """
        super().__init__(requires_hooks)

    @abstractmethod
    def compute(self) -> float:
        """
        Compute the accumulated metric.

        This method must be implemented by any subclass to compute the accumulated
        value of the metric, typically after processing multiple batches.

        Returns:
            float: The computed accumulated metric value.

        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the accumulated state.

        This method must be implemented by any subclass to reset the metric's
        accumulated state.

        """
        pass



import torch
from neurobench.metrics.abstract.workload_metric import AccumulatedMetric
from collections import defaultdict
import numpy as np

try:
    from metavision_ml.metrics.coco_eval import CocoEvaluator
    from metavision_ml.data import box_processing as box_api
    from metavision_sdk_core import EventBbox

    METAVISION_AVAILABLE = True
except ImportError:
    METAVISION_AVAILABLE = False


class CocoMap(AccumulatedMetric):
    """
    COCO mean average precision.

    Measured for event data based on Perot2020, Supplementary B (https://arxiv.org/abs/2009.13436)
        - Skips first 0.5s of each sequence
        - Bounding boxes with diagonal size smaller than 60 pixels are ignored

    """

    def __init__(self):
        """
        Initialize the CocoMap metric.

        Raises:
            ImportError: If the `metavision_ml` and `metavision_sdk_core` packages are not installed on a supported platform.

        """

        if not METAVISION_AVAILABLE:
            raise ImportError(
                "metavision_ml and metavision_sdk_core are required for COCO_mAP metric. "
                "Please install them on a supported platform."
            )

        super().__init__(requires_hooks=False)
        self.dt_detections = defaultdict(list)
        self.gt_detections = defaultdict(list)
        self.evaluator = CocoEvaluator(
            classes=["background"] + ["pedestrian", "two wheeler", "car"],
            height=360,
            width=640,
        )

    def reset(self):
        """
        Reset metric state.

        Clears all accumulated detections and reinitializes the `CocoEvaluator`.

        """
        self.dt_detections = defaultdict(list)
        self.gt_detections = defaultdict(list)
        self.evaluator = CocoEvaluator(
            classes=["background"] + ["pedestrian", "two wheeler", "car"],
            height=360,
            width=640,
        )

    def __call__(self, model, preds, data):
        """
        Accumulate predictions and ground truth detections over batches.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: COCO mean average precision.

        """

        targets = data[1]
        video_infos = data[2]["video_infos"]
        frame_is_labeled = data[2]["frame_is_labeled"]
        skip_us = 500000

        for t in range(len(targets)):
            for i in range(len(targets[t])):
                gt_boxes = targets[t][i]

                pred = preds[t][i]

                video_info, tbin_start, _ = video_infos[i]

                if video_info.padding or frame_is_labeled[t, i] is False:
                    continue

                name = video_info.path
                if name not in self.dt_detections:
                    self.dt_detections[name] = [np.zeros((0), dtype=box_api.EventBbox)]
                if name not in self.gt_detections:
                    self.gt_detections[name] = [np.zeros((0), dtype=box_api.EventBbox)]
                assert video_info.start_ts == 0
                ts = tbin_start + t * video_info.delta_t

                if ts < skip_us:
                    continue

                if isinstance(gt_boxes, torch.Tensor):
                    gt_boxes = gt_boxes.cpu().numpy()
                if gt_boxes.dtype == np.float32:
                    gt_boxes = box_api.box_vectors_to_bboxes(
                        gt_boxes[:, :4], gt_boxes[:, 4], ts=ts
                    )

                if pred["boxes"] is not None and len(pred["boxes"]) > 0:
                    boxes = pred["boxes"].cpu().data.numpy()
                    labels = pred["labels"].cpu().data.numpy()
                    scores = pred["scores"].cpu().data.numpy()
                    dt_boxes = box_api.box_vectors_to_bboxes(
                        boxes, labels, scores, ts=ts
                    )
                    self.dt_detections[name].append(dt_boxes)
                else:
                    self.dt_detections[name].append(np.zeros((0), dtype=EventBbox))

                if len(gt_boxes):
                    gt_boxes["t"] = ts
                    self.gt_detections[name].append(gt_boxes)
                else:
                    self.gt_detections[name].append(np.zeros((0), dtype=EventBbox))

        return 0.0  # too heavy to compute every iteration

    def compute(self):
        """
        Compute COCO mAP using accumulated data.

        Returns:
            float: COCO mean average precision.

        """
        print("Computing COCO mAP.")
        for key in self.gt_detections:
            self.evaluator.partial_eval(
                [np.concatenate(self.gt_detections[key])],
                [np.concatenate(self.dt_detections[key])],
            )
        coco_kpi = self.evaluator.accumulate()
        return coco_kpi["mean_ap"]



import torch


def cylce_tuple(tup):
    """Returns a copy of the tuple with binary elements."""
    tup_copy = []
    for t in tup:
        if isinstance(t, tuple):
            tup_copy.append(cylce_tuple(t))
        elif t is not None:
            t = t.detach().clone()
            t[t != 0] = 1
            tup_copy.append(t)
    return tuple(tup_copy)


def cylce_tuple_ones(tup):
    """Returns a copy of the tuple with ones elements."""
    tup_copy = []
    for t in tup:
        if isinstance(t, tuple):
            tup_copy.append(cylce_tuple(t))
        elif t is not None:
            t = t.detach().clone()
            t[t != 0] = 1
            t[t == 0] = 1
            tup_copy.append(t)
    return tuple(tup_copy)


def binary_inputs(inputs, all_ones=False):
    """Returns a copy of the inputs with binary elements, all ones if all_ones is
    True."""
    in_states = True  # assume that input is tuple of inputs and states. If not, then set to False
    spiking = False

    with torch.no_grad():
        # TODO: should change this code block so that all inputs get cloned
        if isinstance(inputs, tuple):
            # input is first element, rest is hidden states
            test_ins = inputs[0]

            # NOTE: this only checks first input as everything else can be seen as hidden states in rnn block
            if len(test_ins[(test_ins != 0) & (test_ins != 1) & (test_ins != -1)]) == 0:
                spiking = True
            if not all_ones:
                inputs = cylce_tuple(inputs)
            else:
                inputs = cylce_tuple_ones(inputs)
        else:
            # clone tensor since it may be used as input to other layers
            inputs = inputs.detach().clone()
            in_states = False
            if len(inputs[(inputs != 0) & (inputs != 1) & (inputs != -1)]) == 0:
                spiking = True

            inputs[inputs != 0] = 1
            if all_ones:
                inputs[inputs == 0] = 1
    return inputs, spiking, in_states



from neurobench.metrics.abstract.workload_metric import AccumulatedMetric
from neurobench.metrics.utils.layers import single_layer_MACs


class SynapticOperations(AccumulatedMetric):
    """
    Number of synaptic operations.

    This metric computes the number of Multiply-Accumulate operations (MACs) for
    Artificial Neural Networks (ANN) and Accumulation operations (ACs) for Spiking
    Neural Networks (SNN).

    """

    def __init__(self):
        """Initialize SynapticOperations metric."""

        super().__init__(requires_hooks=True)
        self.MAC = 0
        self.AC = 0
        self.total_synops = 0
        self.total_samples = 0

    def reset(self):
        """
        Reset the metric state for a new evaluation.

        Clears all accumulated values for MAC, AC, synaptic operations, and the total
        number of samples.

        """
        self.MAC = 0
        self.AC = 0
        self.total_synops = 0
        self.total_samples = 0

    def __call__(self, model, preds, data):
        """
        Accumulate the Multiply-Accumulate (MAC) operations or Accumulation (AC)
        operations during the forward pass.

        This method accumulates the operations based on the model's connections, and differentiates between
        ANN (MACs) and SNN (ACs) operations based on the spiking activity.


        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
            inputs: A tensor of model inputs.
        Returns:
            float: Multiply-accumulates.

        """
        for hook in model.connection_hooks:
            inputs = hook.inputs  # copy of the inputs, delete hooks after
            for spikes in inputs:
                # spikes is batch, features, see snntorchmodel wrappper
                # for single_in in spikes:
                if len(spikes) == 1:
                    spikes = spikes[0]
                hook.hook.remove()
                operations, spiking = single_layer_MACs(spikes, hook.layer)
                total_ops, _ = single_layer_MACs(spikes, hook.layer, total=True)
                self.total_synops += total_ops
                if spiking:
                    self.AC += operations
                else:
                    self.MAC += operations
                hook.register_hook()
        # ops_per_sample = ops / data[0].size(0)
        self.total_samples += data[0].size(0)
        return self.compute()

    def compute(self):
        """
        Compute the average number of operations per sample.

        Returns:
            dict: A dictionary containing:

                - "Effective_MACs": The average MACs per sample.

                - "Effective_ACs": The average ACs per sample.

                - "Dense": The average total synaptic operations per sample.

        """

        if self.total_samples == 0:
            return {"Effective_MACs": 0, "Effective_ACs": 0, "Dense": 0}
        ac = self.AC / self.total_samples
        mac = self.MAC / self.total_samples
        total_synops = self.total_synops / self.total_samples
        return {"Effective_MACs": mac, "Effective_ACs": ac, "Dense": total_synops}



from torch.utils.data import Dataset
from abc import abstractmethod


class NeuroBenchDataset(Dataset):
    """
    Abstract class for NeuroBench datasets.

    Individual datasets are responsible for collating and splitting data. The top-level
    benchmark expects that the dataset is wrapped into a DataLoader.

    """

    def __init__(self):
        """Init dataset."""
        super().__init__()

    @abstractmethod
    def __len__(self):
        """Returns length of dataset."""
        pass

    @abstractmethod
    def __getitem__(self, idx):
        """Returns a sample from the dataset."""
        pass



from neurobench.blocks.layer import STATELESS_LAYERS, RECURRENT_LAYERS, RECURRENT_CELLS
import torch
from .input import binary_inputs
from .binary_copy import make_binary_copy


def stateless_layer_macs(inputs, layer, total):
    # then multiply the binary layer with the diagonal matrix to get the MACs
    layer_bin = make_binary_copy(layer, all_ones=total)

    # bias is not considered as a synaptic operation
    # in the future you can change this parameter to include bias
    bias = False
    if layer_bin.bias is not None and not bias:
        # suppress the bias to zero
        layer_bin.bias.data = torch.zeros_like(layer_bin.bias.data)

    nr_updates = layer_bin(
        inputs
    )  # this returns the number of MACs for every output neuron: if spiking neurons only AC
    return nr_updates.sum()


def recurrent_layer_macs(inputs, layer, total):
    # layer_bin = make_binary_copy(layer, all_ones=total)
    attribute_names = []
    for i in range(layer.num_layers):
        param_names = ["weight_ih_l{}{}", "weight_hh_l{}{}"]
        if layer.bias:
            param_names += ["bias_ih_l{}{}", "bias_hh_l{}{}"]
        if layer.proj_size > 0:  # it is lstm
            param_names += ["weight_hr_l{}{}"]

        attribute_names += [x.format(i, "") for x in param_names]
        if layer.bidirectional:
            suffix = "_reverse"
            attribute_names += [x.format(i, suffix) for x in param_names]
    raise "This layer is not yet supported by NeuroBench."
    return 0


def lstm_cell_macs(inputs, out_ih, out_hh, layer_bin, biases, in_states):
    # number of operations for lstmcells
    # i = sigmoid(Wii*x + bii + Whi*h + bhi)
    # f = sigmoid(Wif*x + bif + Whf*h + bhf)
    # g = tanh(Wig*x + big + Whg*h + bhg)
    # o = sigmoid(Wio*x + bio + Who*h + bho)

    # c = f*c + i*g
    # h = o*tanh(c)

    # inputs = (x,(h,c))
    if in_states:
        out_hh = torch.matmul(layer_bin.weight_hh, inputs[1][0].transpose(0, -1))

    # out_ih[out_ih!=0] = 1
    # out_hh[out_hh!=0] = 1

    out = out_ih + out_hh

    ifgo_macs = out.sum()  # accounts for i,f,g,o WITHOUT biases

    out += biases  # biases are added here for computation of c and h which depend on correct computation of ifgo
    out[out != 0] = 1

    # out is vector with i,f,g,o, shape is 4*h, batch
    hidden = out.shape[0] // 4
    ifgo = out.reshape(4, hidden, -1)  # 4, h, B
    if in_states:
        # inputs[1][1] shape is [B, h]
        # element-wise multiply (vector products f*c + i*g)
        c_1 = ifgo[1, :] * inputs[1][1].transpose(0, -1) + ifgo[0, :] * ifgo[2, :]
    else:
        c_1 = ifgo[0, :] * ifgo[2, :]

    ifgoc_macs = ifgo_macs + c_1.sum()

    c_1[c_1 != 0] = 1
    output = ifgo[3, :] * c_1  # drop tanh as does not affect 1 vs 0
    output[output != 0] = 1
    return output.sum() + ifgoc_macs


def rnn_cell_macs(inputs, out_ih, out_hh, layer_bin, in_states):
    if in_states:
        out_hh = torch.matmul(layer_bin.weight_hh, inputs[1].transpose(0, -1))
    out = out_ih + out_hh  # no biases for synaptic operations
    return out.sum()


def gru_cell_macs(
    inputs, out_ih, out_hh, bias_ih, bias_hh, layer_bin, biases, in_states
):
    # r = sigmoid(Wir*x + bir + Whr*h + bhr)
    # z = sigmoid(Wiz*x + biz + Whz*h + bhz)
    # n = tanh(Win*x + bin + r*(Whn*h + bhn))
    # h = (1-z)*n + z*h
    macs = 0
    if in_states:
        out_hh = torch.matmul(layer_bin.weight_hh, inputs[1].transpose(0, -1))

    rzn = out_ih + out_hh
    # multiplications of all weights and inputs/hidden states
    # Wir*x, Whr*h, Wiz*x, Whz*h, Win*x, Whn*h
    macs += rzn.sum()  # multiplications of all weights and inputs/hidden states
    rzn += biases  # add biases

    hidden = rzn.shape[0] // 3
    rzn = rzn.reshape(3, hidden, -1)  # 3, h, B
    out_hh = out_hh.reshape(3, hidden, -1)
    bias_hh = bias_hh.reshape(3, hidden, -1)

    out_hh_n = out_hh[2, :] + bias_hh[2, :]
    r = rzn[0, :]  # get r
    z = rzn[1, :]

    r[r != 0] = 1
    out_hh_n[out_hh_n != 0] = 1

    n_hh_term_macs = (
        r * out_hh_n
    )  # elementwise_multiplication to find macs of r*(Whn*h + bhn) specifically
    n_hh_term_macs[n_hh_term_macs != 0] = 1
    macs += n_hh_term_macs.sum()

    # note hh part of n is already binarized, does not influence calculation of macs for n
    n = out_hh[2, :] + bias_ih[2, :] + n_hh_term_macs
    n[n != 0] = 1
    z_a = 1 - z
    # only do this now because affects z_a
    z[z != 0] = 1
    z_a[z_a != 0] = 1
    t_1 = z_a * n
    t_2 = z * inputs[1].transpose(0, -1)  # inputs are shape [B, h], all else is [h, B]

    t_1[t_1 != 0] = 1
    t_2[t_2 != 0] = 1
    out_nrs = t_1 + t_2
    macs += out_nrs.sum()
    return macs


def recurrent_cell_macs(inputs, layer, total, in_states):
    # NOTE: sigmoid and tanh will never change a non-zero value to zero or vice versa
    # NOTE: these activation functions are currently NOT included in NeuroBench
    # if no explicit states are passed to recurrent layers, then h and c are initialized to zero (pytorch convention)
    macs = 0
    layer_bin = make_binary_copy(layer, all_ones=total)
    # layer_weight_ih is [4*hidden_size, input_size]
    # inputs[0].transpose(0, -1) is [input_size, batch_size]
    out_ih = torch.matmul(
        layer_bin.weight_ih, inputs[0].transpose(0, -1)
    )  # accounts for i,f,g,o
    out_hh = torch.zeros_like(out_ih)
    # out shape is 4*h, batch, for hidden feature dim h

    biases = 0
    bias_ih = 0
    bias_hh = 0
    # out matrices are now features, batches
    if layer_bin.bias:
        bias_ih = layer_bin.bias_ih.unsqueeze(0).transpose(0, -1)
        bias_hh = layer_bin.bias_hh.unsqueeze(0).transpose(0, -1)
        biases = bias_ih + bias_hh

    if isinstance(layer, torch.nn.LSTMCell):
        macs = lstm_cell_macs(inputs, out_ih, out_hh, layer_bin, biases, in_states)
    elif isinstance(layer, torch.nn.RNNCell):
        macs = rnn_cell_macs(inputs, out_ih, out_hh, layer_bin, in_states)
    elif isinstance(layer, torch.nn.GRUCell):
        macs = gru_cell_macs(
            inputs, out_ih, out_hh, bias_ih, bias_hh, layer_bin, biases, in_states
        )

    return macs


def single_layer_MACs(inputs, layer, total=False):
    """
    Computes the MACs for a single layer.

    returns effective operations if total=False, else total operations (including zero operations)
    Supported layers: Linear, Conv1d, Conv2d, Conv3d, RNNCellBase, LSTMCell, GRUCell

    """
    macs = 0

    # copy input
    inputs, spiking, in_states = binary_inputs(inputs, all_ones=total)

    if isinstance(layer, STATELESS_LAYERS):
        macs = stateless_layer_macs(inputs, layer, total)
    elif isinstance(layer, RECURRENT_LAYERS):
        macs = recurrent_layer_macs(inputs, layer, total)
    elif isinstance(layer, RECURRENT_CELLS):
        macs = recurrent_cell_macs(inputs, layer, total, in_states)

    return int(macs), spiking



from abc import ABC, abstractmethod
from neurobench.models import NeuroBenchModel


class StaticMetric(ABC):
    """
    Abstract base class for static metrics.

    A static metric computes a value based on the model, and is designed to operate
    without the need for batch processing or iterative accumulation over multiple
    samples. This metric is expected to return a single value when called, which is
    typically computed over the entire model.

    """

    @abstractmethod
    def __call__(self, model: NeuroBenchModel) -> float:
        """
        Compute the static metric for the given model.

        This abstract method must be implemented by any subclass to define
        how the metric should be computed based on the model.

        Args:
            model (NeuroBenchModel): The model whose performance or properties the metric evaluates.

        Returns:
            float: The computed value of the metric for the model.

        Notes:
            - The method should return a single float value representing the
              result of the metric computation. It does not require batch
              processing or tracking multiple samples.
            - This is intended for metrics that compute a single, static
              evaluation of a model, such as those based on its architecture,
              weights, or overall performance.

        """
        pass



from neurobench.metrics.abstract import StaticMetric


class ParameterCount(StaticMetric):
    """A metric that counts the number of parameters in a model."""

    def __call__(self, model):
        """
        Count the number of parameters in a model.

        Args:
            model: A NeuroBenchModel.
        Returns:
            int: Number of parameters in the model.

        """
        return sum(p.numel() for p in model.__net__().parameters())



from .neurobench_model import NeuroBenchModel


class TorchModel(NeuroBenchModel):
    """The TorchModel class wraps an nn.Module."""

    def __init__(self, net):
        """
        Initializes the TorchModel class.

        Args:
            net: A PyTorch nn.Module.

        """
        super().__init__()

        self.net = net
        self.net.eval()

    def __call__(self, batch):
        """
        Wraps forward pass of torch.nn model.

        Args:
            batch: A PyTorch tensor of shape (batch, timesteps, features*)

        Returns:
            preds: either a tensor to be compared with targets or passed to
                NeuroBenchPostProcessors.

        """
        return self.net(batch)

    def __net__(self):
        """Returns the underlying network."""
        return self.net



"""
Speech2Spikes License Copyright © 2023 Accenture.

Speech2Spikes is made available under a proprietary license that permits using,
copying, sharing, and making derivative works from Speech2Spikes and its source
code for academics/non-commercial purposes only, as long as the above copyright
notice and this permission notice are included in all copies of the software.

All distribution of Speech2Spikes in any form (source or executable), including
any derivative works that you create or to which you contribute, must be under
the terms of this license. You must inform recipients that any form of
Speech2Spikes and its derivatives is governed by the terms of this license, and
how they can obtain a copy of this license and a copy of the source code of
Speech2Spikes. You may not attempt to alter or restrict the recipients’ rights
in any form. If you are interested to use Speech2Spikes and/or develop
derivatives for commercial purposes, licenses can be purchased from Accenture,
please contact neuromorphic_inquiries@accenture.com for more information.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

You agree to indemnify and hold Accenture harmless from and against all
liabilities, claims and suits and to pay all costs and expenses thereby
incurred, including reasonable legal fees and court courts, arising out of,
caused by or in any way connected with your use of Speech2Spikes.

The original code can be found at:
https://github.com/Accenture/speech2spikes

"""

from neurobench.processors.abstract.preprocessor import NeuroBenchPreProcessor
import torch
import torchaudio


def tensor_to_events(batch, threshold=1, device=None):
    """Converts a batch of continuous signals to binary spikes via delta modulation
    (https://en.wikipedia.org/wiki/Delta_modulation).

    Args:
        batch (Tensor): PyTorch tensor of shape (..., timesteps).
        threshold (float): The difference between the residual and signal that
            will be considered an increase or decrease. Defaults to 1.
        device (torch.device, optional): A torch.Device used by PyTorch for the
            computation. Defaults to None.

    Returns:
        Tensor: A PyTorch int8 tensor of events of shape (..., timesteps).

    TODO:
        Add support for using multiple channels for polarity instead of signs.
    """
    events = torch.zeros(batch.shape)
    levels = torch.round(batch[..., 0])
    if device:
        events = events.to(device)

    for t in range(batch.shape[-1]):
        events[..., t] = (batch[..., t] - levels > threshold).to(torch.int8) - (
            batch[..., t] - levels < -threshold
        ).to(torch.int8)
        levels += events[..., t] * threshold
    return events


class S2SPreProcessor(NeuroBenchPreProcessor):
    """The SpikeEncoder class manages the conversion from raw audio into spikes and
    stores the required conversion parameters."""

    def __init__(self, device=None, transpose=True, log_offset=1e-6):
        """
        Args:
            device (torch.device, optional): A torch.Device used by PyTorch for the
                computation. Defaults to None.
            transpose (bool, optional): Whether to transpose the input tensor before processing.
                If the input tensor is of shape (batch, channels, timesteps), this should be true.
            log_offset (float, optional): A small value added to the MelSpectrogram before log is applied
        """
        self.device = device
        self.transpose = transpose
        self.log_offset = log_offset
        self.spec_kwargs = {
            "sample_rate": 16000,
            "n_mels": 20,
            "n_fft": 512,
            "f_min": 20,
            "f_max": 4000,
            "hop_length": 80,
        }
        self.threshold = 1
        self.transform = torchaudio.transforms.MelSpectrogram(**self.spec_kwargs).to(
            device
        )

    def __call__(self, batch):
        """Converts raw audio data to spikes using Speech2Spikes algorithm
        (https://doi.org/10.1145/3584954.3584995)

        Args:
            batch: A tuple of data and corresponding targets (data_tensor, targets)

        Returns:
            tensors: PyTorch int8 tensor of shape (batch, timesteps, ...)
            targets: A tensor of corresponding targets.

        TODO:
            Add support for cumulative sum of features
        """
        timesteps = batch[0].shape[1]

        tensors = batch[0]
        targets = batch[1]
        if len(batch) == 3:
            kwargs = batch[2]
        else:
            kwargs = None

        # Tensors will be batch, timestep, channels and need to be transposed

        if self.transpose:
            tensors = tensors.transpose(1, 2)
        tensors = self.transform(tensors)
        if self.log_offset:
            tensors = tensors + self.log_offset
        tensors = torch.log(tensors)
        tensors = tensor_to_events(
            tensors, threshold=self.threshold, device=self.device
        )
        tensors = tensors.transpose(
            1, 3
        ).squeeze()  # Transpose back to batch, timestep, channel

        # torchaudio seems to return one extra timestep, get rid of the zero timestep
        if tensors.shape[1] == (timesteps / self.spec_kwargs["hop_length"] + 1):
            tensors = tensors[:, 1:, :]

        if kwargs:
            return tensors, targets, kwargs
        else:
            return tensors, targets

    def configure(self, threshold=1, **spec_kwargs):
        """
        Allows the user to configure parameters of the S2S class and the MelSpectrogram
        transform from torchaudio.

        Go to (https://pytorch.org/audio/main/generated/torchaudio.transforms.MelSpectrogram.html)
        for more information on the available transform parameters.

        Args:
            threshold (float): The difference between the residual and signal that
                will be considered an increase or decrease. Defaults to 1.
            **spec_kwargs: Keyword arguments passed to torchaudio's MelSpectrogram.

        """
        self.threshold = threshold

        self.spec_kwargs = {**self.spec_kwargs, **spec_kwargs}
        self.transform = torchaudio.transforms.MelSpectrogram(**self.spec_kwargs).to(
            self.device
        )



import torch
from neurobench.metrics.abstract.workload_metric import AccumulatedMetric
from collections import defaultdict


class MembraneUpdates(AccumulatedMetric):
    """
    Membrane potential updates metric.

    This metric computes the number of membrane potential updates occurring during the
    forward pass of the model. The updates are tracked per neuron, per layer.

    """

    def __init__(self):
        """Initialize the MembraneUpdates metric."""
        super().__init__(requires_hooks=True)
        self.total_samples = 0
        self.neuron_membrane_updates = defaultdict(int)

    def reset(self):
        """Reset the metric state for a new evaluation."""
        self.total_samples = 0
        self.neuron_membrane_updates = defaultdict(int)

    def __call__(self, model, preds, data):
        """
        Accumulate the number of membrane updates for each model forward pass.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: Number of membrane potential updates.

        """
        for hook in model.activation_hooks:
            layer_type = hook.layer.__class__.__name__
            updates = self.neuron_membrane_updates[layer_type]

            # Vectorized computation of updates
            if len(hook.pre_fire_mem_potential) > 1:
                pre_fire_mem = torch.stack(hook.pre_fire_mem_potential[1:])
                post_fire_mem = torch.stack(hook.post_fire_mem_potential[1:])
                updates += torch.count_nonzero(pre_fire_mem - post_fire_mem).item()

            # Add the number of elements in the first post_fire_mem_potential
            if hook.post_fire_mem_potential:
                updates += hook.post_fire_mem_potential[0].numel()

            # Update the dictionary
            self.neuron_membrane_updates[layer_type] = updates

        # Increment total_samples
        self.total_samples += data[0].size(0)

        # Return computed results
        return self.compute()

    def compute(self):
        """
        Compute the total membrane updates normalized by the number of samples.

        Returns:
            float: Compute the total updates to each neuron's membrane potential within the model,
            aggregated across all neurons and normalized by the number of samples processed.

        """
        if self.total_samples == 0:
            return 0

        total_mem_updates = sum(self.neuron_membrane_updates.values())

        return total_mem_updates / self.total_samples



from neurobench.processors.abstract.postprocessor import NeuroBenchPostProcessor


class Aggregate(NeuroBenchPostProcessor):
    """Returns aggregated spikes."""

    def __call__(self, spikes):
        """
        Returns the aggregated spikes.

        Args:
            spikes (Tensor): A torch tensor of spikes of shape (batch, timestep, classes)

        Returns:
            Tensor: A torch tensor of spikes of shape (batch, classes)

        """
        return spikes.sum(1)



import torch
from torch import Tensor
from neurobench.metrics.utils.decorators import check_shapes

from neurobench.metrics.abstract.workload_metric import WorkloadMetric


class SMAPE(WorkloadMetric):
    """Symmetric mean absolute percentage error of the model predictions."""

    def __init__(self):
        """Initialize the SMAPE metric."""

        super().__init__(requires_hooks=False)

    @check_shapes
    def __call__(self, model, preds: Tensor, data: Tensor) -> float:
        """
        Compute symmetric mean absolute percentage error.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: Symmetric mean absolute percentage error.

        """

        smape = 200 * torch.mean(
            torch.abs(preds - data[1]) / (torch.abs(preds) + torch.abs(data[1]))
        )
        return torch.nan_to_num(smape, nan=200.0).item()



import sys
from contextlib import redirect_stdout
from os import mkdir
from tqdm import tqdm
from snntorch import utils
from neurobench.metrics.manager.static_manager import StaticMetricManager
from neurobench.metrics.manager.workload_manager import WorkloadMetricManager
from neurobench.processors.manager import ProcessorManager
from neurobench.models import NeuroBenchModel, SNNTorchModel
from torch.utils.data import DataLoader
from neurobench.processors.abstract import (
    NeuroBenchPreProcessor,
    NeuroBenchPostProcessor,
)
from neurobench.metrics.abstract import StaticMetric, WorkloadMetric
import json
import csv
import os
from typing import Literal, List, Type, Optional, Dict, Any, Callable, Tuple
import pathlib
import snntorch
from torch import Tensor
import torch


class Benchmark:
    """Top-level benchmark class for running benchmarks."""

    def __del__(self):
        if hasattr(self, "workload_metric_manager"):
            self.workload_metric_manager.cleanup_hooks(self.model)

    def __init__(
        self,
        model: NeuroBenchModel,
        dataloader: Optional[DataLoader],
        preprocessors: Optional[
            List[
                NeuroBenchPreProcessor
                | Callable[[Tuple[Tensor, Tensor]], Tuple[Tensor, Tensor]]
            ]
        ],
        postprocessors: Optional[
            List[NeuroBenchPostProcessor | Callable[[Tensor], Tensor]]
        ],
        metric_list: List[List[Type[StaticMetric | WorkloadMetric]]],
    ):
        """
        Args:
            model: A NeuroBenchModel.
            dataloader: A PyTorch DataLoader.
            preprocessors: A list of NeuroBenchPreProcessors or callable functions (e.g. lambda) with matching interfaces.
            postprocessors: A list of NeuroBenchPostProcessors or callable functions (e.g. lambda) with matching interfaces.
            metric_list: A list of lists of StaticMetric and WorkloadMetric classes of metrics to run.
                First item is StaticMetrics, second item is WorkloadMetrics.
        """

        self.model = model
        self.dataloader = dataloader  # dataloader not dataset
        self.processor_manager = ProcessorManager(preprocessors, postprocessors)
        self.static_metric_manager = StaticMetricManager(metric_list[0])
        self.workload_metric_manager = WorkloadMetricManager(metric_list[1])
        self.workload_metric_manager.register_hooks(model)
        self.results = None

    def run(
        self,
        quiet: bool = False,
        verbose: bool = False,
        dataloader: Optional[DataLoader] = None,
        preprocessors: Optional[
            NeuroBenchPreProcessor
            | Callable[[Tuple[Tensor, Tensor]], Tuple[Tensor, Tensor]]
        ] = None,
        postprocessors: Optional[
            NeuroBenchPostProcessor | Callable[[Tensor], Tensor]
        ] = None,
        device: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Runs batched evaluation of the benchmark.

        Args:
            dataloader (Optional): override DataLoader for this run.
            preprocessors (Optional): override preprocessors for this run.
            postprocessors (Optional): override postprocessors for this run.
            quiet (bool, default=False): If True, output is suppressed.
            verbose (bool, default=False): If True, metrics for each bach will be printed.
                                           If False (default), metrics are accumulated and printed after all batches are processed.
            device (Optional): use device for this run (e.g. 'cuda' or 'cpu').

        Returns:
            Dict[str, Any]: A dictionary of results.

        """
        with redirect_stdout(None if quiet else sys.stdout):
            print("Running benchmark")

            self.results = None
            results = self.static_metric_manager.run_metrics(self.model)

            dataloader = dataloader if dataloader is not None else self.dataloader

            if preprocessors is not None:
                self.processor_manager.replace_preprocessors(preprocessors)
            if postprocessors is not None:
                self.processor_manager.replace_postprocessors(postprocessors)

            self.workload_metric_manager.initialize_metrics()

            dataset_len = len(dataloader.dataset)

            if device is not None:
                self.model.__net__().to(device)

            batch_num = 0
            for data in tqdm(dataloader, total=len(dataloader), disable=quiet):
                # convert data to tuple
                data = tuple(data) if not isinstance(data, tuple) else data

                if device is not None:
                    data = (data[0].to(device), data[1].to(device))

                batch_size = data[0].size(0)

                # Preprocessing data
                input, target = self.processor_manager.preprocess(data)

                # Run model on test data
                preds = self.model(input)

                # Postprocessing data
                preds = self.processor_manager.postprocess(preds)

                # Data metrics
                batch_results = self.workload_metric_manager.run_metrics(
                    self.model, preds, data, batch_size, dataset_len
                )
                self.workload_metric_manager.reset_hooks(self.model)

                if verbose:
                    results.update(batch_results)
                    print(f"\nBatch num {batch_num + 1}/{len(dataloader)}")
                    print(dict(results))

                batch_num += 1

            results.update(self.workload_metric_manager.results)
            self.workload_metric_manager.clean_results()
            self.results = dict(results)
        return self.results

    def save_benchmark_results(
        self, file_path: str, file_format: Literal["json", "csv", "txt"] = "json"
    ) -> None:
        """
        Save benchmark results to a specified file in the chosen format.

        Args:
            file_path (str):
                Path to the output file (excluding the extension). The method
                automatically appends the appropriate extension based on the
                chosen file format.
            file_format (Literal["json", "csv", "txt"], default="json"):
                The format in which the results should be saved. Supported formats:

                - `"json"`: Saves the results as a JSON file with formatted indentation.

                - `"csv"`: Saves the results as a CSV file with keys as headers and values as the first row.

                - `"txt"`: Saves the results as a plain text file with one key-value pair per line.

        Raises:
            ValueError:
                If the provided `file_format` is not one of the supported formats (`"json"`, `"csv"`, `"txt"`).

        """
        file_format = file_format.lower()

        # Ensure the directory exists
        pathlib.Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)

        # JSON format
        if file_format == "json":
            with open(f"{file_path}.json", "w") as json_file:
                json.dump(self.results, json_file, indent=4)
            print(f"Results saved to {file_path}.json")

        # CSV format
        elif file_format == "csv":
            with open(f"{file_path}.csv", "w", newline="") as csv_file:
                writer = csv.writer(csv_file)

                # Write header (keys)
                writer.writerow(self.results.keys())

                # Write values
                writer.writerow(self.results.values())
            print(f"Results saved to {file_path}.csv")

        # Plain Text format
        elif file_format == "txt":
            with open(f"{file_path}.txt", "w") as txt_file:
                for key, value in self.results.items():
                    txt_file.write(f"{key}: {value}\n")
            print(f"Results saved to {file_path}.txt")

        else:
            raise ValueError(f"Unsupported file format: {file_format}")

    def to_nir(self, dummy_input: Tensor, filename: str, **kwargs) -> None:
        """
        Exports the model to the NIR (Neural Intermediate Representation) format.

        Args:
            dummy_input (torch.Tensor):
                A sample input tensor that matches the input shape of the model.
                This is required for tracing the model during export.
            filename (str):
                The file path where the exported NIR file will be saved.
            **kwargs:
                Additional keyword arguments passed to the `export_to_nir` function
                for customization during the export process.

        Raises:
            ValueError:
                If the installed version of `snntorch` is less than `0.9.0`.

        """
        try:
            import nir
        except ImportError:
            raise ImportError(
                "Exporting to NIR requires the `nir` package. Install it using `pip install nir`."
            )
        if snntorch.__version__ < "0.9.0":
            raise ValueError("Exporting to NIR requires snntorch version >= 0.9.0")

        if snntorch.__version__ >= "0.9.0":
            from snntorch.export_nir import export_to_nir

        nir_graph = export_to_nir(self.model.__net__(), dummy_input, **kwargs)
        nir.write(filename, nir_graph)
        print(f"Model exported to {filename}")

    def to_onnx(self, dummy_input: Tensor, filename: str, **kwargs) -> None:
        """
        Exports the model to the ONNX (Open Neural Network Exchange) format.

        Args:
            dummy_input (torch.Tensor):
                A sample input tensor that matches the input shape of the model.
                This tensor is required for tracing the model during the export process.
            filename (str):
                The file path where the ONNX model will be saved, including the `.onnx` extension.
            **kwargs:
                Additional keyword arguments passed to the `torch.onnx.export` function
                for customization during the export process.

        Raises:
            RuntimeError:
                If an error occurs during the ONNX export process.

        """
        if dummy_input.requires_grad:
            dummy_input = dummy_input.detach()

        for param in self.model.__net__().parameters():
            param.requires_grad = False

        if isinstance(self.model, SNNTorchModel):
            utils.reset(self.model.__net__())

        for buffer in self.model.__net__().buffers():
            buffer.requires_grad = False

        with torch.no_grad():
            self.model.__net__().eval()
            torch.onnx.export(self.model.__net__(), dummy_input, filename, **kwargs)
        print(f"Model exported to {filename}")



from functools import wraps


def check_shapes(func):
    """Decorator to check that the shapes of predictions and labels are the same before
    executing the decorated function."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Assuming `preds` and `labels` are the first two positional arguments
        if len(args) >= 2:
            preds, data = args[2:]
            if preds.shape != data[1].shape:
                raise ValueError(
                    f"In {func.__name__}: 'preds' and 'labels' must have the same shape. "
                    f"Got preds.shape={preds.shape} and data[1].shape={data[1].shape}."
                )
        return func(*args, **kwargs)

    return wrapper



from .dataset import *
from .utils import *
from ..utils import _lazy_import


def SpeechCommands(*args, **kwargs):
    return _lazy_import("neurobench.datasets", ".speech_commands", "SpeechCommands")(
        *args, **kwargs
    )


def PrimateReaching(*args, **kwargs):
    return _lazy_import("neurobench.datasets", ".primate_reaching", "PrimateReaching")(
        *args, **kwargs
    )


def Gen4DetectionDataLoader(*args, **kwargs):
    return _lazy_import(
        "neurobench.datasets", ".megapixel_automotive", "Gen4DetectionDataLoader"
    )(*args, **kwargs)


def MackeyGlass(*args, **kwargs):
    return _lazy_import("neurobench.datasets", ".mackey_glass", "MackeyGlass")(
        *args, **kwargs
    )


def WISDM(*args, **kwargs):
    return _lazy_import("neurobench.datasets", ".WISDM", "WISDM")(*args, **kwargs)


def MSWC(*args, **kwargs):
    return _lazy_import("neurobench.datasets", ".MSWC_dataset", "MSWC")(*args, **kwargs)



from neurobench.metrics.abstract import StaticMetric


class Footprint(StaticMetric):
    """A metric that counts the memory footprint of a model."""

    def __call__(self, model):
        """
        Count the memory footprint of a model.

        Args:
            model: A NeuroBenchModel.
        Returns:
            float: Memory footprint of the model.

        """
        param_size = 0
        for param in model.__net__().parameters():
            param_size += param.numel() * param.element_size()

        buffer_size = 0
        for buffer in model.__net__().buffers():
            buffer_size += buffer.numel() * buffer.element_size()

        return param_size + buffer_size



from neurobench.metrics.abstract.workload_metric import AccumulatedMetric
from collections import defaultdict
import torch


class ActivationSparsityByLayer(AccumulatedMetric):
    """
    Sparsity layer-wise of model activations.

    Calculated as the number of zero activations over the number of activations layer by
    layer, over all timesteps, samples in data.

    """

    def __init__(self):
        """Initialize the ActivationSparsityByLayer metric."""
        self.layer_sparsity = defaultdict(float)
        self.layer_neuro_num = defaultdict(int)
        self.layer_spike_num = defaultdict(int)

    def reset(self):
        """Reset the metric."""
        self.layer_sparsity = defaultdict(float)
        self.layer_neuro_num = defaultdict(int)
        self.layer_spike_num = defaultdict(int)

    def __call__(self, model, preds, data):
        """
        Compute activation sparsity layer by layer.

        Args:
            model: A NeuroBenchModel.
            preds: A tensor of model predictions.
            data: A tuple of data and labels.
        Returns:
            float: Activation sparsity

        """

        for hook in model.activation_hooks:
            name = hook.name
            if name is None:
                continue
            for output in hook.activation_outputs:
                spike_num, neuro_num = torch.count_nonzero(
                    output.dequantize() if output.is_quantized else output
                ).item(), torch.numel(output)

                self.layer_spike_num[name] += spike_num
                self.layer_neuro_num[name] += neuro_num

        return self.compute()

    def compute(self):
        """Compute the activation sparsity layer by layer."""
        for key in self.layer_neuro_num:
            sparsity = (
                (self.layer_neuro_num[key] - self.layer_spike_num[key])
                / self.layer_neuro_num[key]
                if self.layer_neuro_num[key] != 0
                else 0.0
            )
            self.layer_sparsity[key] = sparsity
        return dict(self.layer_sparsity)



import snntorch as snn
from abc import ABC


class NeuronHook(ABC):
    """
    Hook class for an activation layer in a NeuroBenchModel.

    Output of the activation layer in each forward pass will be stored.

    """

    def __init__(self, layer, name=None):
        """
        Initializes the class.

        A forward hook is registered for the activation layer.

        Args:
            layer: The activation layer which is a PyTorch nn.Module.

        """
        self.activation_outputs = []
        self.activation_inputs = []
        self.pre_fire_mem_potential = []
        self.post_fire_mem_potential = []
        self.name = name
        if layer is not None:
            self.hook = layer.register_forward_hook(self.hook_fn)
            self.hook_pre = layer.register_forward_pre_hook(self.pre_hook_fn)
        else:
            self.hook = None
            self.hook_pre = None

        self.layer = layer  # the activation layer

        # Check if the layer is a spiking layer (SpikingNeuron is the superclass of all snnTorch spiking layers)
        self.spiking = isinstance(layer, snn.SpikingNeuron)

    def pre_hook_fn(self, layer, input):
        """
        Hook function that will be called before each forward pass of the activation
        layer.

        Each input of the activation layer will be stored.

        Args:
            layer: The registered layer
            input: Input of the registered layer

        """
        self.activation_inputs.append(input)
        if self.spiking and hasattr(layer, "mem"):
            self.pre_fire_mem_potential.append(layer.mem)

    def hook_fn(self, layer, input, output):
        """
        Hook function that will be called after each forward pass of the activation
        layer.

        Each output of the activation layer will be stored.

        Args:
            layer: The registered layer
            input: Input of the registered layer
            output: Output of the registered layer

        """
        if self.spiking:
            self.activation_outputs.append(output[0])
            if hasattr(layer, "mem"):
                self.post_fire_mem_potential.append(layer.mem)

        else:
            self.activation_outputs.append(output)

    # def empty_hook(self):
    #     """Deletes the contents of the hooks, but keeps the hook registered."""
    #     self.activation_outputs = []
    #     self.activation_inputs = []

    def reset(self):
        """Resets the stored activation outputs and inputs."""
        self.activation_outputs.clear()
        self.activation_inputs.clear()
        self.pre_fire_mem_potential.clear()
        self.post_fire_mem_potential.clear()

    def close(self):
        """Remove the registered hook."""
        if self.hook:
            self.hook.remove()
        if self.hook_pre:
            self.hook_pre.remove()



import torch.nn as nn

STATELESS_LAYERS = (
    nn.Linear,
    nn.Conv2d,
    nn.Conv1d,
    nn.Conv3d,
)

RECURRENT_CELLS = (nn.RNNCellBase,)

RECURRENT_LAYERS = (nn.RNNBase,)


SUPPORTED_LAYERS = STATELESS_LAYERS + RECURRENT_LAYERS + RECURRENT_CELLS



def convert_to_class_name(metric_name):
    """
    Convert a metric name from a string to a class name.

    Args:
        metric_name (str): The metric name in snake_case or other formatting.

    Returns:
        str: The corresponding class name in CamelCase.

    """
    # Convert snake_case to CamelCase
    if metric_name in ["sMAPE", "r2", "mse"]:
        return metric_name.upper()
    return "".join(word.title() for word in metric_name.split("_"))
